import os
import sys
import time
import logging
from alita.base import BaseExceptionHandler, BaseStaticHandler
from alita.exceptions import default_exceptions, NotFound, BadRequest,\
    InternalServerError
from alita.response import FileResponse


class ExceptionHandler(BaseExceptionHandler):
    _status_handlers = None
    _exception_handlers = None

    def __init__(self, app=None, logger=None):
        super(ExceptionHandler, self).__init__(app)
        self._status_handlers = {}
        self._exception_handlers = {}
        self.logger = logger or logging.getLogger(__name__)

    def add_status_handler(self, code, handler):
        assert isinstance(code, int)
        self._status_handlers[code] = handler

    def add_exception_handler(self, exception_class, handler):
        assert issubclass(exception_class, Exception)
        self._exception_handlers[exception_class] = handler

    def _lookup_exception_handler(self, exc):
        for cls in type(exc).__mro__:
            if cls in self._exception_handlers:
                return self._exception_handlers[cls]
        return None

    def ruder_error_response(self, request, exc):
        exc_type, exc_value, tb = sys.exc_info()
        self.app.log_exception(request, (exc_type, exc_value, tb))
        # TODO: 此处可以渲染一个默认的500报错页面，当然也可以通过注册500的handler函数来定义渲染
        error_content = str(exc)
        return InternalServerError()

    def default_handler(self, request, exc):
        try:
            exc_class = type(exc)
            if issubclass(exc_class, int):
                return default_exceptions[exc]()
            elif issubclass(exc_class, self.app.exception_class):
                return exc
            else:
                assert issubclass(exc_class, Exception)
                raise exc
        except Exception as ex:
            try:
                return self.ruder_error_response(request, ex)
            except Exception as exc:
                return InternalServerError(str(exc))

    async def process_exception(self, request, exc):
        handler = None
        try:
            if isinstance(exc, self.app.exception_class):
                handler = self._status_handlers.get(exc.code)
            if handler is None:
                handler = self._lookup_exception_handler(exc)
            if handler is None:
                handler = self.default_handler
            return await self.app.get_awaitable_result(handler, request, exc)
        except Exception as ex:
            return await self.process_exception(request, ex)


class StaticHandler(BaseStaticHandler):
    async def send_static_file(self, request, file_name):
        if not self.has_static_folder:
            raise RuntimeError('No static folder for this object')
        file_max_age = self.send_file_max_age
        filename = os.path.join(self.static_folder, file_name)
        try:
            if not os.path.isfile(filename):
                raise NotFound()
        except (TypeError, ValueError):
            raise BadRequest()
        headers = {}
        if file_max_age is not None:
            headers['Cache-Control'] = 'max-age=%s' % file_max_age
            headers['Expires'] = time.time() + file_max_age
        return FileResponse(filename, headers=headers)



import re
import attr
from enum import Enum
from alita.base import BaseRoute, BaseRouter
from alita.converters import CONVERTER_TYPES
from urllib.parse import urljoin
from alita.utils import set_query_parameter, get_request_url
from alita.exceptions import NotFound, BadRequest, InternalServerError


class NoMatchFound(NotFound):
    pass


class Match(Enum):
    NONE = 0
    PARTIAL = 1
    FULL = 2


def replace_params(path, param_converters, path_params):
    for key, value in list(path_params.items()):
        if "{" + key + "}" in path:
            converter = param_converters[key]
            value = converter.to_string(value)
            path = path.replace("{" + key + "}", value)
            path_params.pop(key)
    for key, value in path_params.items():
        set_query_parameter(path, key, value)
    return path


# Match parameters in URL paths, eg. '<param_name:int/float/str/path>'
PARAM_REGEX = re.compile("<([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?>")


def compile_path(path):
    path_regex = "^"
    path_format = ""

    idx = 0
    param_converters = {}
    for match in PARAM_REGEX.finditer(path):
        param_name, converter_type = match.groups("str")
        converter_type = converter_type.lstrip(":")
        if converter_type not in CONVERTER_TYPES:
            raise RuntimeError(f"Unknown path converter '{converter_type}'")
        converter = CONVERTER_TYPES[converter_type]
        path_regex += path[idx: match.start()]
        path_regex += f"(?P<{param_name}>{converter.regex})"
        path_format += path[idx: match.start()]
        path_format += "{%s}" % param_name
        param_converters[param_name] = converter
        idx = match.end()

    path_regex += path[idx:] + "$"
    path_format += path[idx:]
    return re.compile(path_regex), path_format, param_converters


class URLPath:
    def __init__(self, path, protocol="http", host="", params=None):
        self.path = path
        self.protocol = protocol
        self.host = host
        self.params = params or {}
        assert protocol in ("http", "websocket")

    def make_url(self, base_url=None, request=None):
        if request is not None:
            assert not base_url, 'Cannot set both "base_url" and "request".'
            return get_request_url(request)
        elif base_url:
            assert not request, 'Cannot set both "request" and "**components".'
            return urljoin(base_url, self.path)
        else:
            raise ValueError('base_url or request need supply one.')


@attr.s
class RouteMatch:
    status = attr.ib(validator=attr.validators.instance_of(Match))
    endpoint = attr.ib(validator=attr.validators.instance_of(str))
    view_func = attr.ib()
    path_params = attr.ib(validator=attr.validators.instance_of(dict))


class Route(BaseRoute):
    def __init__(self, path, endpoint, view_func, methods=None):
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.endpoint = endpoint
        self.view_func = view_func
        methods = methods or getattr(view_func, 'methods', None) or ('GET',)
        self.methods = set([method.upper() for method in methods])
        if "GET" in self.methods:
            self.methods |= set(["HEAD"])

        self.path_regex, self.path_format, self.param_converters = compile_path(path)

    def match(self, request):
        if request.scheme == "http":
            match = self.path_regex.match(request.path)
            if not match:
                raise NoMatchFound()
            matched_params = match.groupdict()
            for key, value in matched_params.items():
                matched_params[key] = self.param_converters[key].convert(value)
            path_params = dict(request.path_params, **matched_params)
            if self.methods and request.method not in self.methods:
                status = Match.PARTIAL
            else:
                status = Match.FULL
            return RouteMatch(
                status,
                self.endpoint,
                self.view_func,
                path_params,
            )
        else:
            raise NoMatchFound()

    def url_path_for(self, endpoint, **path_params) -> URLPath:
        seen_params = set(path_params.keys())
        expected_params = set(self.param_converters.keys())
        if endpoint != self.endpoint or seen_params != expected_params:
            raise NoMatchFound()
        path = replace_params(self.path_format, self.param_converters, path_params)
        return URLPath(path=path, protocol="http")

    def __eq__(self, other):
        return (
            isinstance(other, Route)
            and self.path == other.path
            and self.endpoint == other.endpoint
            and self.methods == other.methods
        )


class Router(BaseRouter):
    def __init__(self, app, routes=None):
        self.app = app
        self.routes = routes or []

    def add_route(self, path, endpoint, view_func, methods=None):
        route = Route(
            path,
            endpoint=endpoint,
            view_func=view_func,
            methods=methods,
        )
        self.routes.append(route)

    def match(self, request):
        for route in self.routes:
            try:
                route_math = route.match(request)
                if route_math.status == Match.PARTIAL:
                    raise BadRequest()
                return route_math
            except NoMatchFound:
                pass
        raise NoMatchFound()

    def url_path_for(self, name: str, **path_params: str) -> URLPath:
        for route in self.routes:
            try:
                return route.url_path_for(name, **path_params)
            except NoMatchFound:
                pass
        raise NoMatchFound()



from alita.base import BaseFactory


class AppFactory(BaseFactory):
    def __init__(self, app=None):
        self.app = app

    def create_request_object(self, environ):
        from alita.request import Request
        return Request(self.app, environ)

    def create_base_response_class(self):
        from alita.response import HTTPResponse
        return HTTPResponse

    def create_base_exception_class(self):
        from alita.exceptions import HTTPException
        return HTTPException

    def create_exception_handler_object(self):
        from alita.handler import ExceptionHandler
        return ExceptionHandler(self.app)

    def create_router_object(self):
        from alita.routing import Router
        return Router(self.app)

    def create_static_handler(self):
        from alita.handler import StaticHandler
        return StaticHandler(self.app)



# -*- coding: utf-8 -*-
signals_available = False
try:
    from blinker import Namespace
    signals_available = True
except ImportError:
    class Namespace(object):
        def signal(self, name, doc=None):
            return _FakeSignal(name, doc)

    class _FakeSignal(object):
        def __init__(self, name, doc=None):
            self.name = name
            self.__doc__ = doc

        def _fail(self, *args, **kwargs):
            raise RuntimeError('signalling support is unavailable '
                               'because the blinker library is '
                               'not installed.')
        send = lambda *a, **kw: None
        connect = disconnect = has_receivers_for = receivers_for = \
            temporarily_connected_to = connected_to = _fail
        del _fail

_signals = Namespace()

request_started = _signals.signal('request-started')
request_finished = _signals.signal('request-finished')
got_request_exception = _signals.signal('got-request-exception')
message_flashed = _signals.signal('message-flashed')



import os
import logging
import httptools
import signal
import asyncio
import functools
from datetime import datetime
from alita.serve.utils import *
from urllib.parse import unquote

HIGH_WATER_LIMIT = 65536


class ServiceUnavailable:
    def __init__(self, body=None, status=503, headers=None, content_type="text/plain"):
        self.content_type = content_type
        self.body = self._encode_body(body)
        self.status = status
        self._cookies = None
        self.headers = headers or {}

    def _encode_body(self, data):
        try:
            return data.encode()
        except AttributeError:
            return str(data or "").encode()

    async def __call__(self, environ, on_response):
        on_response(self)

    def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        # This is all returned in a kind-of funky way
        # We tried to make this as fast as possible in pure python
        body, timeout_header = b"", b""
        if keep_alive and keep_alive_timeout is not None:
            timeout_header = b"Keep-Alive: %d\r\n" % keep_alive_timeout
        self.headers["Content-Type"] = self.headers.get(
            "Content-Type", self.content_type
        )
        headers = self._parse_headers()
        description = b'Service Unavailable'

        return (
            b"HTTP/%b %d %b\r\n" b"Connection: %b\r\n" b"%b" b"%b\r\n" b"%b"
        ) % (
            version.encode(),
            self.status,
            description,
            b"keep-alive" if keep_alive else b"close",
            timeout_header,
            headers,
            body,
        )

    def _parse_headers(self):
        headers = b""
        for name, value in self.headers.items():
            try:
                headers += b"%b: %b\r\n" % (
                    name.encode(),
                    value.encode("utf-8"),
                )
            except AttributeError:
                headers += b"%b: %b\r\n" % (
                    str(name).encode(),
                    str(value).encode("utf-8"),
                )
        return headers


class HttpProtocol(asyncio.Protocol):
    DEFAULT_TYPE = "http"
    DEFAULT_VERSION = "1.1"

    def __init__(self, app, config, server_state):
        self.app = app
        self.config = config
        self.loop = config.loop
        self.logger = config.logger
        self.access_log = config.access_log and (self.logger.level <= logging.INFO)
        self.parser = httptools.HttpRequestParser(self)
        self.ws_protocol = config.ws_protocol
        self.root_path = config.root_path
        self.limit_concurrency = config.limit_concurrency
        self.keep_alive_timeout = config.keep_alive_timeout

        # Timeouts
        self.timeout_keep_alive_task = None
        self.timeout_keep_alive = config.timeout_keep_alive

        # Global state
        self.server_state = server_state
        self.connections = server_state.connections
        self.tasks = server_state.tasks
        self.default_headers = server_state.default_headers + config.default_headers

        # Per-connection state
        self.transport = None
        self.server = None
        self.client = None
        self.scheme = None
        self.pipeline = []

        # Per-request state
        self.url = None
        self.environ = None
        self.body = b""
        self.more_body = True
        self.headers = []
        self.expect_100_continue = False
        self.message_event = asyncio.Event()
        self.message_event.set()

    # Protocol interface
    def connection_made(self, transport):
        self.connections.add(self)
        self.transport = transport
        self.server = get_local_addr(transport)
        self.client = get_remote_addr(transport)
        self.scheme = "https" if is_ssl(transport) else "http"

        if self.logger.level <= logging.DEBUG:
            self.logger.debug("%s - Connected", self.client)

    def connection_lost(self, exc):
        self.connections.discard(self)
        if self.logger.level <= logging.DEBUG:
            self.logger.debug("%s - Disconnected", self.client)
        self.message_event.set()

    def cancel_timeout_keep_alive_task(self):
        if self.timeout_keep_alive_task is not None:
            self.timeout_keep_alive_task.cancel()
            self.timeout_keep_alive_task = None

    def data_received(self, data):
        self.cancel_timeout_keep_alive_task()
        try:
            self.parser.feed_data(data)
        except httptools.parser.errors.HttpParserError as exc:
            msg = "Invalid HTTP request received."
            self.logger.warning(msg)
            self.transport.close()
        except httptools.HttpParserUpgrade as exc:
            self.handle_upgrade()

    def handle_upgrade(self):
        upgrade_value = None
        for name, value in self.headers:
            if name == b"upgrade":
                upgrade_value = value.lower()

        if upgrade_value != b"websocket" or self.ws_protocol is None:
            msg = "Unsupported upgrade request."
            self.logger.warning(msg)
            content = [STATUS_TEXT[400]]
            for name, value in self.default_headers:
                content.extend([name, b": ", value, b"\r\n"])
            content.extend(
                [
                    b"content-type: text/plain; charset=utf-8\r\n",
                    b"content-length: " + str(len(msg)).encode("ascii") + b"\r\n",
                    b"connection: close\r\n",
                    b"\r\n",
                    msg.encode("ascii"),
                ]
            )
            self.transport.write(b"".join(content))
            self.transport.close()
            return

        self.connections.discard(self)
        method = self.environ["method"].encode()
        output = [method, b" ", self.url, b" HTTP/1.1\r\n"]
        for name, value in self.environ["headers"]:
            output += [name, b": ", value, b"\r\n"]
        output.append(b"\r\n")
        protocol = self.ws_protocol(
            config=self.config, server_state=self.server_state
        )
        protocol.connection_made(self.transport)
        protocol.data_received(b"".join(output))
        self.transport.set_protocol(protocol)

    # Parser callbacks
    def on_url(self, url):
        method = self.parser.get_method()
        parsed_url = httptools.parse_url(url)
        path = parsed_url.path.decode("ascii")
        if "%" in path:
            path = unquote(path)
        self.url = url
        self.expect_100_continue = False
        self.environ = {
            "url": url.decode(),
            "parsed_url": parsed_url,
            "type": self.DEFAULT_TYPE,
            "http_version": self.DEFAULT_VERSION,
            "server": self.server,
            "client": self.client,
            "scheme": self.scheme,
            "host": self.server[0],
            "port": self.server[1],
            "method": method.decode("ascii"),
            "root_path": self.root_path,
            "path": path,
            "query_string": (parsed_url.query if parsed_url.query else b"").decode(),
            "headers": self.headers,
            "default_headers": self.default_headers,
        }

    def on_header(self, name: bytes, value: bytes):
        name = name.lower()
        if name == b"expect" and value.lower() == b"100-continue":
            self.expect_100_continue = True
        self.headers.append((name, value))

    def on_headers_complete(self):
        http_version = self.parser.get_http_version()
        if http_version != self.DEFAULT_VERSION:
            self.environ["http_version"] = http_version
        if self.parser.should_upgrade():
            return
        self.environ.update(
            protocol=self,
            transport=self.transport,
            logger=self.logger,
            access_log=self.access_log,
            expect_100_continue=self.expect_100_continue,
            keep_alive=http_version != "1.0",
            keep_alive_timeout=self.keep_alive_timeout
        )
    
    def on_body(self, body: bytes):
        if self.parser.should_upgrade():
            return
        self.body += body
        if len(self.body) > HIGH_WATER_LIMIT:
            self.transport.pause_reading()
        self.message_event.set()

    def on_message_complete(self):
        if self.parser.should_upgrade():
            return
        self.more_body = False
        self.message_event.set()
        self.cancel_timeout_keep_alive_task()
        self.environ.update(body=self.body.decode())
        self.process_request()

    def log_response(self, response):
        if self.access_log:
            self.logger.info('[access] %s - - [%s] "%s %s" %s -',
                             self.environ['host'],
                             datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                             self.environ['method'],
                             self.environ['path'],
                             response.status)

    def process_request(self):
        # Standard case - start processing the request.
        # Handle 503 responses when 'limit_concurrency' is exceeded.
        if self.limit_concurrency is not None and (
                len(self.connections) >= self.limit_concurrency
                or len(self.tasks) >= self.limit_concurrency
        ):
            app = ServiceUnavailable()
            message = "Exceeded concurrency limit."
            self.logger.warning(message)
        else:
            app = self.app
        task = self.loop.create_task(app(self.environ, self.on_response))
        task.add_done_callback(self.tasks.discard)
        self.tasks.add(task)

    async def on_response(self, response):
        # Callback for pipelined HTTP requests to be started.
        self.server_state.total_requests += 1
        response.set_protocol(self)
        output_content = await response.output(
            self.environ["http_version"],
            self.environ["keep_alive"],
            self.environ["keep_alive_timeout"]
        )
        self.transport.write(output_content)
        self.log_response(response)

        if not self.transport.is_closing():
            self.transport.close()
            self.transport = None
        else:
            # Set a short Keep-Alive timeout.
            self.timeout_keep_alive_task = self.loop.call_later(
                self.timeout_keep_alive, self.timeout_keep_alive_handler
            )

    def shutdown(self):
        """
        Called by the server to commence a graceful shutdown.
        """
        if not self.transport.is_closing():
            self.transport.close()

    def pause_writing(self):
        """
        Called by the transport when the write buffer exceeds the high water mark.
        """
        self.message_event.clear()

    def resume_writing(self):
        """
        Called by the transport when the write buffer drops below the low water mark.
        """
        self.message_event.set()

    async def drain(self):
        await self.message_event.wait()

    def timeout_keep_alive_handler(self):
        """
        Called on a keep-alive connection if no new data is received after a short delay.
        """
        self.shutdown()

    def push_data(self, data):
        self.transport.write(data)


class ServerState:
    """
    Shared servers state that is available between all protocol instances.
    """

    def __init__(self, total_requests=0, connections=None, tasks=None, default_headers=None):
        self.total_requests = total_requests
        self.connections = connections or set()
        self.tasks = tasks or set()
        self.default_headers = default_headers or []


class Server(object):
    def __init__(self, app, config, server_state=None):
        self.app = app
        self.config = config
        self.started = False
        self.loop = config.loop
        self.logger = config.logger
        self.servers = []
        self.server_state = server_state or ServerState()

    def install_signal_handlers(self):
        try:
            for sig in (signal.SIGINT, signal.SIGTERM):
                self.loop.add_signal_handler(sig, self.loop.stop)
        except NotImplementedError:
            self.logger.warning("loop.add_signal_handler not implemented on this platform.")

    def run(self):
        server = functools.partial(
            self.config.http_protocol,
            app=self.app,
            config=self.config,
            server_state=self.server_state
        )
        asyncio_server_kwargs = (
            self.config.asyncio_server_kwargs if self.config.asyncio_server_kwargs else {}
        )
        server_coroutine = self.loop.create_server(
            server,
            self.config.host,
            self.config.port,
            ssl=self.config.ssl,
            reuse_port=self.config.reuse_port,
            sock=self.config.socket,
            backlog=self.config.backlog,
            **asyncio_server_kwargs
        )

        try:
            http_server = self.loop.run_until_complete(server_coroutine)
        except BaseException:
            self.logger.exception("Unable to start server")
            return
        self.install_signal_handlers()
        pid = os.getpid()
        try:
            self.started = True
            self.servers = [server]
            self.logger.info("Starting worker [%s]", pid)
            message = "Server running on http://%s:%d (Press CTRL+C to quit)"
            self.logger.info(message % (self.config.host, self.config.port))
            self.loop.run_forever()
        finally:
            self.logger.info("Stopping worker [%s]", pid)
            # Wait for event loop to finish and all connections to drain
            http_server.close()
            self.loop.run_until_complete(http_server.wait_closed())

            # Complete all tasks on the loop
            for connection in self.server_state.connections:
                connection.close_if_idle()

            # Gracefully shutdown timeout.
            # We should provide graceful_shutdown_timeout,
            # instead of letting connection hangs forever.
            # Let's roughly calcucate time.
            start_shutdown = 0
            while self.server_state.connections and (
                    start_shutdown < self.config.graceful_shutdown_timeout):
                self.loop.run_until_complete(asyncio.sleep(0.1))
                start_shutdown = start_shutdown + 0.1

            # Force close non-idle connection after waiting for
            # graceful_shutdown_timeout
            coros = []
            for conn in self.server_state.connections:
                if hasattr(conn, "websocket") and conn.websocket:
                    coros.append(conn.websocket.close_connection())
                else:
                    conn.close()

            _shutdown = asyncio.gather(*coros, loop=self.loop)
            self.loop.run_until_complete(_shutdown)
            self.loop.close()


__all__ = [
    "HttpProtocol",
    "Server",
    "ServerState"
]



import os
import re
import sys
import json
import pkgutil
import importlib
from urllib.parse import urlencode, parse_qs, urlsplit,\
    urlunsplit, unquote_to_bytes

_ENTITY_HEADERS = frozenset(
    [
        "allow",
        "content-encoding",
        "content-language",
        "content-length",
        "content-location",
        "content-md5",
        "content-range",
        "content-type",
        "expires",
        "last-modified",
        "extension-header",
    ]
)
_etag_re = re.compile(r'([Ww]/)?(?:"(.*?)"|(.*?))(?:\s*,\s*|$)')
_unsafe_header_chars = set('()<>@,;:\"/[]?={} \t')
_option_header_piece_re = re.compile(r'''
    ;\s*
    (?P<key>
        "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
    |
        [^\s;,=*]+  # token
    )
    \s*
    (?:  # optionally followed by =value
        (?:  # equals sign, possibly with encoding
            \*\s*=\s*  # * indicates extended notation
            (?P<encoding>[^\s]+?)
            '(?P<language>[^\s]*?)'
        |
            =\s*  # basic notation
        )
        (?P<value>
            "[^"\\]*(?:\\.[^"\\]*)*"  # quoted string
        |
            [^;,]+  # token
        )?
    )?
    \s*
''', flags=re.VERBOSE)
_option_header_start_mime_type = re.compile(r',\s*([^;,\s]+)([;,]\s*.+)?')


def escape(s, quote=None):
    """Replace special characters "&", "<", ">" and (") to HTML-safe sequences.

    There is a special handling for `None` which escapes to an empty string.

    .. versionchanged:: 0.9
       `quote` is now implicitly on.

    :param s: the string to escape.
    :param quote: ignored.
    """
    if s is None:
        return ''
    elif hasattr(s, '__html__'):
        return str(s.__html__())
    elif not isinstance(s, str):
        s = str(s)
    if quote is not None:
        from warnings import warn
        warn(DeprecationWarning('quote parameter is implicit now'), stacklevel=2)
    s = s.replace('&', '&amp;').replace('<', '&lt;') \
        .replace('>', '&gt;').replace('"', "&quot;")
    return s


def to_unicode(x, charset=sys.getdefaultencoding(), errors='strict',
               allow_none_charset=False):
    if x is None:
        return None
    if not isinstance(x, bytes):
        return str(x)
    if charset is None and allow_none_charset:
        return x
    return x.decode(charset, errors)


class cached_property(object):
    """Cached property descriptor.

    Caches the return value of the get method on first call.

    Examples:
        .. code-block:: python

            @cached_property
            def connection(self):
                return Connection()

            @connection.setter  # Prepares stored value
            def connection(self, value):
                if value is None:
                    raise TypeError('Connection must be a connection')
                return value

            @connection.deleter
            def connection(self, value):
                # Additional action to do at del(self.attr)
                if value is not None:
                    print('Connection {0!r} deleted'.format(value)
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.__get = fget
        self.__set = fset
        self.__del = fdel
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        try:
            return obj.__dict__[self.__name__]
        except KeyError:
            value = obj.__dict__[self.__name__] = self.__get(obj)
            return value

    def __set__(self, obj, value):
        if obj is None:
            return self
        if self.__set is not None:
            value = self.__set(obj, value)
        obj.__dict__[self.__name__] = value

    def __delete__(self, obj, _sentinel=object()):
        if obj is None:
            return self
        value = obj.__dict__.pop(self.__name__, _sentinel)
        if self.__del is not None and value is not _sentinel:
            self.__del(obj, value)

    def setter(self, fset):
        return self.__class__(self.__get, fset, self.__del)

    def deleter(self, fdel):
        return self.__class__(self.__get, self.__set, fdel)


class ImportFromStringError(Exception):
    pass


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        if not isinstance(dotted_path, str):
            return dotted_path
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        ) from err


def import_from_string(import_str):
    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = (
            'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError as exc:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(
            message.format(attrs_str=attrs_str, module_str=module_str)
        )

    return instance


def get_root_path(import_name):
    """Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.

    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, '__file__'):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == '__main__':
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, 'get_filename'):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, '__file__', None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError('No root path can be found for the provided '
                               'module "%s".  This can happen because the '
                               'module came from an import hook that does '
                               'not provide file name information or because '
                               'it\'s a namespace package.  In this case '
                               'the root path needs to be explicitly '
                               'provided.' % import_name)

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def check_serialize(data):
    try:
        json.dumps(data)
        return True
    except:
        return False


def has_message_body(status):
    """
    According to the following RFC message body and length SHOULD NOT
    be included in responses status 1XX, 204 and 304.
    https://tools.ietf.org/html/rfc2616#section-4.4
    https://tools.ietf.org/html/rfc2616#section-4.3
    """
    return status not in (204, 304) and not (100 <= status < 200)


def is_entity_header(header):
    """Checks if the given header is an Entity Header"""
    return header.lower() in _ENTITY_HEADERS


def remove_entity_headers(headers, allowed=("content-location", "expires")):
    """
    Removes all the entity headers present in the headers given.
    According to RFC 2616 Section 10.3.5,
    Content-Location and Expires are allowed as for the
    "strong cache validator".
    https://tools.ietf.org/html/rfc2616#section-10.3.5

    returns the headers without the entity headers
    """
    allowed = set([h.lower() for h in allowed])
    headers = {
        header: value
        for header, value in headers.items()
        if not is_entity_header(header) or header.lower() in allowed
    }
    return headers


def set_query_parameter(url, param_name, param_value):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    query_params[param_name] = [param_value]
    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def get_request_url(request, root_only=False, strip_querystring=False):
    scheme = request.scheme
    path = request.root_path + request.path
    query_string = request.query_string
    host_header = None
    if root_only:
        path, query_string = '/', ''
    for key, value in request.headers.items():
        if key == 'host':
            host_header = value
            break

    if host_header:
        url = f"{scheme}://{host_header}{path}"
    else:
        host, port = request.server
        default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
        if port == default_port:
            url = f"{scheme}://{host}{path}"
        else:
            url = f"{scheme}://{host}:{port}{path}"
    if query_string and not strip_querystring:
        url += "?" + query_string
    return url


def parse_options_header(value, multiple=False):
    if not value:
        return '', {}
    result = []
    value = "," + value.replace("\n", ",")

    def unquote_header_value(val, is_filename=False):
        if val and val[0] == val[-1] == '"':
            val = val[1:-1]
            if not is_filename or val[:2] != '\\\\':
                return val.replace('\\\\', '\\').replace('\\"', '"')
        return val

    while value:
        match = _option_header_start_mime_type.match(value)
        if not match:
            break
        result.append(match.group(1))  # mimetype
        options = {}
        # Parse options
        rest = match.group(2)
        while rest:
            optmatch = _option_header_piece_re.match(rest)
            if not optmatch:
                break
            option, encoding, _, option_value = optmatch.groups()
            option = unquote_header_value(option)
            if option_value is not None:
                option_value = unquote_header_value(
                    option_value,
                    option == 'filename')
                if encoding is not None:
                    option_value = unquote_to_bytes(option_value).decode(encoding)
            options[option] = option_value
            rest = rest[optmatch.end():]
        result.append(options)
        if multiple is False:
            return tuple(result)
        value = rest

    return tuple(result) if result else ('', {})


def total_seconds(td):
    """Returns the total seconds from a timedelta object.

    :param timedelta td: the timedelta to be converted in seconds

    :returns: number of seconds
    :rtype: int
    """
    return td.days * 60 * 60 * 24 + td.seconds



import http


def get_remote_addr(transport):
    info = transport.get_extra_info("peername")
    if info is not None and isinstance(info, (list, tuple)) and len(info) == 2:
        return (str(info[0]), int(info[1]))
    return None


def get_local_addr(transport):
    info = transport.get_extra_info("sockname")
    if info is not None and isinstance(info, (list, tuple)) and len(info) == 2:
        return (str(info[0]), int(info[1]))
    return None


def is_ssl(transport):
    return bool(transport.get_extra_info("sslcontext"))


def get_status_text(status_code):
    try:
        phrase = http.HTTPStatus(status_code).phrase.encode()
    except ValueError as exc:
        phrase = b""
    return phrase


STATUS_TEXT = {
    status_code: get_status_text(status_code) for status_code in range(100, 600)
}



import asyncio
import logging
from alita.serve.server import HttpProtocol
from alita.serve.ws import WebSocketProtocol


def get_logger(log_level):
    logging.basicConfig(format="%(message)s", level=log_level)
    logger = logging.getLogger("uvicorn")
    logger.setLevel(log_level)
    return logger


def init_loop(uvloop=True):
    if uvloop:
        try:
            import uvloop
            asyncio.get_event_loop().close()
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except ImportError:
            pass
    return asyncio.get_event_loop()


class ServerConfig:
    DEFAULT_PROTOCOL = HttpProtocol

    def __init__(
        self,
        host="127.0.0.1",
        port=8000,
        uds=None,
        fd=None,
        ssl=None,
        socket=None,
        connections=None,
        loop=None,
        http_protocol=None,
        ws=None,
        log_level=logging.INFO,
        logger=None,
        uvloop=True,
        access_log=True,
        wsgi=False,
        debug=False,
        backlog=100,
        proxy_headers=False,
        root_path="",
        limit_concurrency=None,
        limit_max_requests=None,
        timeout_keep_alive=5,
        timeout_notify=30,
        callback_notify=None,
        reuse_port=False,
        install_signal_handlers=True,
        asyncio_server_kwargs=None,
        graceful_shutdown_timeout=10.0,
        keep_alive_timeout=5,
        default_headers=None
    ):
        self.host = host
        self.port = port
        self.uds = uds
        self.fd = fd
        self.ssl = ssl
        self.socket = socket
        self.connections = connections
        self.loop = loop or init_loop(uvloop)
        self.log_level = log_level
        self.logger = logger or get_logger(self.log_level)
        self.access_log = access_log
        self.wsgi = wsgi
        self.debug = debug
        self.backlog = backlog
        self.proxy_headers = proxy_headers
        self.root_path = root_path
        self.limit_concurrency = limit_concurrency
        self.limit_max_requests = limit_max_requests
        self.timeout_keep_alive = timeout_keep_alive
        self.timeout_notify = timeout_notify
        self.callback_notify = callback_notify
        self.reuse_port = reuse_port
        self.asyncio_server_kwargs = asyncio_server_kwargs
        self.graceful_shutdown_timeout = graceful_shutdown_timeout
        self.http_protocol = http_protocol or self.DEFAULT_PROTOCOL
        self.ws_protocol = ws or WebSocketProtocol
        self.keep_alive_timeout = keep_alive_timeout
        self.default_headers = default_headers or []


__all__ = [
    "ServerConfig",
    "init_loop",
]



# -*- coding: utf-8 -*-
import json
import types
import errno
from alita.utils import import_string


class ConfigAttribute(object):
    def __init__(self, name, get_converter=None):
        self.__name__ = name
        self.get_converter = get_converter

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        rv = obj.config[self.__name__]
        if self.get_converter is not None:
            rv = self.get_converter(rv)
        return rv

    def __set__(self, obj, value):
        obj.config[self.__name__] = value


class Config(dict):
    def __init__(self, defaults=None):
        dict.__init__(self, defaults or {})

    def from_pyfile(self, filename, silent=False):
        d = types.ModuleType('config')
        d.__file__ = filename
        try:
            with open(filename, mode='rb') as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (
                errno.ENOENT, errno.EISDIR, errno.ENOTDIR
            ):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        if isinstance(obj, str):
            obj = import_string(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def from_json(self, filename, silent=False):
        try:
            with open(filename) as json_file:
                obj = json.loads(json_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj)

    def from_mapping(self, *mapping, **kwargs):
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(mapping[0].items())
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        mappings.append(kwargs.items())
        for mapping in mappings:
            for (key, value) in mapping:
                if key.isupper():
                    self[key] = value
        return True

    def get_namespace(self, namespace, lowercase=True, trim_namespace=True):
        rv = {}
        for k, v in self.items():
            if not k.startswith(namespace):
                continue
            if trim_namespace:
                key = k[len(namespace):]
            else:
                key = k
            if lowercase:
                key = key.lower()
            rv[key] = v
        return rv

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, dict.__repr__(self))



import os
import six
import sys
import inspect
import logging
import warnings
import itertools
import traceback
import alita.signals as signals
from alita.serve import *
from inspect import isawaitable
from alita.datastructures import ImmutableDict
from alita.config import Config, ConfigAttribute
from alita.factory import AppFactory
from alita.utils import import_string, ImportFromStringError, check_serialize
from alita.base import BaseMiddleware
from alita.response import HtmlResponse, JsonResponse
from alita.exceptions import InternalServerError


class Alita(object):
    config_class = Config
    _default_factory_class = AppFactory
    _view_middleware = []

    app_factory_class = None
    response_class = None
    exception_class = None

    key_prefix = ConfigAttribute('SESSION_KEY_PREFIX')
    session_use_signer = ConfigAttribute('SESSION_USE_SIGNER')
    session_engine = ConfigAttribute('SESSION_ENGINE')
    session_cookie_name = ConfigAttribute('SESSION_COOKIE_NAME')
    session_cookie_path = ConfigAttribute('SESSION_COOKIE_PATH')
    session_cookie_domain = ConfigAttribute('SESSION_COOKIE_DOMAIN')
    session_save_every_request = ConfigAttribute('SESSION_SAVE_EVERY_REQUEST')
    session_cookie_expire = ConfigAttribute('SESSION_COOKIE_EXPIRE')
    session_engine_config = ConfigAttribute('SESSION_ENGINE_CONFIG')
    session_table_name = ConfigAttribute('SESSION_TABLE_NAME')
    send_file_max_age = ConfigAttribute('SEND_FILE_MAX_AGE')

    default_config = ImmutableDict({
        'DEBUG': False,
        'PROPAGATE_EXCEPTIONS': None,
        'PRESERVE_CONTEXT_ON_EXCEPTION': None,
        'SECRET_KEY': None,
        'SESSION_COOKIE_EXPIRE': 30 * 24 * 60 * 60,
        'USE_X_SENDFILE': False,
        'SERVER_NAME': None,
        'APPLICATION_ROOT': '/',
        'SESSION_COOKIE_NAME': 'sessionid',
        'SESSION_COOKIE_DOMAIN': None,
        'SESSION_COOKIE_PATH': None,
        'SESSION_COOKIE_HTTPONLY': True,
        'SESSION_COOKIE_SECURE': False,
        'SESSION_COOKIE_SAMESITE': None,
        'SESSION_KEY_PREFIX': '',
        'SESSION_USE_SIGNER': False,
        'SESSION_REFRESH_EACH_REQUEST': True,
        'MAX_CONTENT_LENGTH': None,
        'SEND_FILE_MAX_AGE': 12 * 60 * 60,
        'TRAP_BAD_REQUEST_ERRORS': None,
        'TRAP_HTTP_EXCEPTIONS': False,
        'EXPLAIN_TEMPLATE_LOADING': False,
        'PREFERRED_URL_SCHEME': 'http',
        'JSON_AS_ASCII': True,
        'JSON_SORT_KEYS': True,
        'JSONIFY_PRETTYPRINT_REGULAR': False,
        'JSONIFY_MIMETYPE': 'application/json',
        'TEMPLATES_AUTO_RELOAD': None,
        'MAX_COOKIE_SIZE': 4093,
        'SESSION_SAVE_EVERY_REQUEST': False,
        'SESSION_EXPIRE_AT_BROWSER_CLOSE': False,
        'SESSION_MUST_SAVE': True,
        'SESSION_TABLE_NAME': 'session',
        "SESSION_ENGINE": "alita.contrib.sessions.backends.mysql",
        # 'SESSION_ENGINE_CONFIG': {
        #     'db': 'gezi',
        #     'host': '192.168.4.5',
        # },
        'SESSION_ENGINE_CONFIG': {
            'host': '192.168.4.5',
            'port': 3306,
            'username': 'root',
            'password': 'tianfu',
            'database': 'gezi'
        },
        'MIDDLEWARE': [
            "alita.contrib.sessions.middleware.SessionMiddleware"
        ],
    })

    def __init__(self, name=None, subdomain_matching=False, static_folder=None,
                 static_url_path=None):
        self.name = name
        self.view_functions = {}
        self.static_folder = static_folder
        self.static_url_path = static_url_path
        self.subdomain_matching = subdomain_matching
        self.config = None
        self.is_running = False
        self.make_config()

        self.before_request_funcs = {}
        self.after_request_funcs = {}
        self.error_handler_spec = {}
        self.blueprints = {}
        self.logger = logging.getLogger(__name__)
        self.load_middleware()

        self.app_factory_class = import_string(self.config.get(
            "APP_FACTORY_CLASS", self._default_factory_class))
        self.app_factory = self.app_factory_class(self)
        self.exception_handler = None
        self.static_handler = None
        self.router = None
        self.make_factory()

    @property
    def debug(self):
        return self.config['DEBUG']

    def load_middleware(self):
        """
        Populate middleware lists from settings.MIDDLEWARE.

        Must be called after the environment is fixed (see __call__ in subclasses).
        """
        self._view_middleware = []
        for middleware_path in reversed(self.config['MIDDLEWARE']):
            try:
                middleware_class = import_string(middleware_path)
                if not issubclass(middleware_class, BaseMiddleware):
                    raise TypeError('Middleware class must implement by '
                                    'BaseMiddleware: %r', middleware_path)
                self._view_middleware.append(middleware_class(self))
            except ImportFromStringError as exc:
                self.logger.debug('Middleware import error: %r', middleware_path)
                raise exc

    def register_middleware(self, middleware_class):
        if middleware_class in self._view_middleware:
            raise Exception("register middleware repeated!")
        self._view_middleware.append(middleware_class(self))

    async def process_request_middleware(self, request):
        for middleware in self._view_middleware:
            await middleware.process_request(request)

    async def process_response_middleware(self, request, response):
        for middleware in self._view_middleware:
            response = await middleware.process_response(request, response)
        return response

    def before_request(self, f):
        self.before_request_funcs.setdefault(None, []).append(f)
        return f

    def after_request(self, f):
        self.after_request_funcs.setdefault(None, []).append(f)
        return f

    def make_config(self):
        defaults = dict(self.default_config)
        self.config = self.config_class(defaults)

    def config_from_py(self, filename, silent=False):
        self.config.from_pyfile(filename, silent)

    def config_from_json(self, filename, silent=False):
        self.config.from_json(filename, silent)

    def config_from_object(self, filename, silent=False):
        self.config.from_object(filename, silent)

    def make_factory(self):
        self.response_class = self.app_factory.create_base_response_class()
        self.exception_class = self.app_factory.create_base_exception_class()
        self.exception_handler = self.app_factory.create_exception_handler_object()
        self.router = self.app_factory.create_router_object()
        self.static_handler = self.app_factory.create_static_handler()

    def error_handler(self, code_or_exception):
        def decorator(func):
            self.register_error_handler(code_or_exception, func)
            return func
        return decorator

    def register_error_handler(self, code_or_exception, func):
        if isinstance(code_or_exception, int):
            self.exception_handler.add_status_handler(code_or_exception, func)
        elif isinstance(code_or_exception, Exception):
            self.exception_handler.add_exception_handler(code_or_exception, func)
        else:
            raise ValueError("decorator error handler code or exception "
                             "is invalid int or exception type!")

    @staticmethod
    def get_endpoint_from_view_func(view_func):
        return view_func.__name__

    def add_url_rule(self, rule, view_func, endpoint=None, methods=None):
        if endpoint is None:
            endpoint = self.get_endpoint_from_view_func(view_func)
        if methods is None:
            methods = getattr(view_func, 'methods', None) or ('GET',)
        elif isinstance(methods, str):
            methods = [methods]
        assert isinstance(methods, (tuple, list))
        methods = set(item.upper() for item in methods)
        self.router.add_route(rule, endpoint, view_func, methods)
        old_func = self.view_functions.get(endpoint)
        if old_func is not None and old_func != view_func:
            raise AssertionError('View function mapping is overwriting an '
                                 'existing endpoint function: %s' % endpoint)
        if not inspect.iscoroutinefunction(view_func):
            warnings.warn("View function [%s] should be async function" % view_func.__name__)
        self.view_functions[endpoint] = view_func

    def route(self, rule, **options):
        def decorator(f):
            self.add_url_rule(rule, f, **options)
            return f
        return decorator

    def run(self, **kwargs):
        config = ServerConfig(**kwargs)
        server = Server(self, config=config)
        server.run()

    async def preprocess_request(self, request):
        bp = request.blueprint
        funcs = self.before_request_funcs.get(None, [])
        if bp is not None and bp in self.before_request_funcs:
            funcs = itertools.chain(funcs, self.before_request_funcs[bp])
        for func in funcs:
            rv = self.get_awaitable_result(func)
            if rv is not None:
                return rv

    async def dispatch_request(self, request):
        if request.routing_exception is not None:
            raise request.routing_exception
        route_match = request.route_match
        return await self.get_awaitable_result(
            route_match.view_func,
            request, **route_match.path_params
        )

    async def make_response(self, response):
        return response

    async def get_awaitable_result(self, func, *args, **kwargs):
        func_result = func(*args, **kwargs)
        if isawaitable(func_result):
            return await func_result
        else:
            return func_result

    async def process_response(self, request, response):
        bp = request.blueprint
        funcs = self.after_request_funcs.get(None, [])
        if bp is not None and bp in self.after_request_funcs:
            funcs = itertools.chain(funcs, self.after_request_funcs[bp])
        for func in funcs:
            response = self.get_awaitable_result(func)
        return response

    async def finalize_response(self, response):
        if isinstance(response, self.response_class):
            pass
        elif isinstance(response, self.exception_class):
            return response.get_response()
        elif isinstance(response, six.string_types):
            return HtmlResponse(response)
        elif check_serialize(response):
            return JsonResponse(response)
        else:
            raise Exception("Response Object Type invalid!")
        return response

    async def finalize_request(self, request, response, from_error_handler=False):
        response = await self.make_response(response)
        try:
            response = await self.process_response(request, response)
            signals.request_finished.send(self, response=response)
        except Exception as ex:
            if not from_error_handler:
                raise ex
            self.logger.exception('Request finalizing failed with an '
                                  'error while handling an error')
        return response

    async def make_dispatch_request_exception(self, exp):
        self.logger.error(str(exp))
        self.logger.exception(traceback.format_exc())
        return InternalServerError(description=traceback.format_exc())

    async def full_dispatch_request(self, request):
        try:
            signals.request_started.send(self)
            await self.process_request_middleware(request)
            response = await self.preprocess_request(request)
            if response is None:
                response = await self.dispatch_request(request)
            response = await self.finalize_request(request, response)
            response = await self.finalize_response(response)
            return await self.process_response_middleware(request, response)
        except Exception as ex:
            exception = await self.exception_handler.process_exception(request, ex)
            raise exception

    async def create_request(self, environ):
        return self.app_factory.create_request_object(environ)

    async def __call__(self, environ, on_response):
        request, response = None, None
        try:
            request = await self.create_request(environ)
            response = await self.full_dispatch_request(request)
        except Exception as ex:
            exception = await self.exception_handler.process_exception(request, ex)
            if isinstance(exception, self.exception_class):
                response = exception.get_response()
            elif isinstance(exception, self.response_class):
                response = exception
            else:
                response = None
        if response:
            await on_response(response)
        else:
            message = "Caught handled exception, response object empty."
            self.logger.error(message)
            await on_response(self.exception_handler.ruder_error_response(
                request, RuntimeError(message)))

    def log_exception(self, request, exc_info):
        self.logger.error('Exception on %s [%s]' % (
            request.path,
            request.method
        ), exc_info=exc_info)

    def register_blueprint(self, blueprint, **options):
        if blueprint.name in self.blueprints:
            raise RuntimeError("blueprint %s has registered to app. "
                               "not allow repeat register." % blueprint.name)
        else:
            self.blueprints[blueprint.name] = blueprint
        blueprint.register(self, **options)

    def url_for(self, request, endpoint, **path_params):
        url_path = self.router.url_path_for(endpoint, **path_params)
        return url_path.make_url(request=request)



import os
import mimetypes
from functools import partial
from urllib.parse import quote_plus
from alita.base import BaseResponse
from aiofiles import open as open_async

try:
    from ujson import dumps as json_dumps
except:
    from json import dumps
    json_dumps = partial(dumps, separators=(",", ":"))


class HTTPResponse(BaseResponse):
    pass


class RawResponse(HTTPResponse):
    """
    Returns response object without encoding the body.
    """
    def __init__(self, body, status=200, headers=None, content_type="application/octet-stream"):
        super(RawResponse, self).__init__(body, status, headers, content_type)


class HtmlResponse(HTTPResponse):
    """
    Returns response object with body in html format.
    """
    def __init__(self, body, status=200, headers=None, content_type="text/html; charset=utf-8"):
        super(HtmlResponse, self).__init__(body, status, headers, content_type)


class TextResponse(HTTPResponse):
    """
    Returns response object with body in text format.
    """
    def __init__(self, body, status=200, headers=None, content_type="text/plain; charset=utf-8"):
        super(TextResponse, self).__init__(body, status, headers, content_type)


class JsonResponse(HTTPResponse):
    """
    Returns response object with body in json format.
    """
    def __init__(self, body, status=200, headers=None, content_type="application/json"):
        super(JsonResponse, self).__init__(body, status, headers, content_type)


class RedirectResponse(HTTPResponse):
    """
    Returns response object with body in json format.
    """
    def __init__(self, to, status=200, headers=None, content_type="text/html; charset=utf-8"):
        headers = headers or {}
        safe_to = quote_plus(to, safe=":/%#?&=@[]!$&'()*+,;")
        headers["Location"] = safe_to
        super(RedirectResponse, self).__init__(None, status, headers, content_type)


class StreamHTTPResponse(BaseResponse):
    def __init__(self, stream_fn, status=200, headers=None, content_type="text/plain"):
        self.stream_fn = stream_fn
        super(StreamHTTPResponse, self).__init__('', status, headers, content_type)

    async def write(self, data):
        data = self._encode_body(data)
        self._protocol.push_data(b"%x\r\n%b\r\n" % (len(data), data))
        await self._protocol.drain()

    async def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        if not self.has_protocol():
            raise RuntimeError("Http protocol not set, "
                               "stream response can not execute.")
        self.headers["Transfer-Encoding"] = "chunked"
        self.headers.pop("Content-Length", None)
        headers = super(StreamHTTPResponse, self).output(
            version, keep_alive, keep_alive_timeout)
        self._protocol.push_data(headers)
        await self._protocol.drain()
        await self.stream_fn(self)
        self._protocol.push_data(b"0\r\n\r\n")


class FileResponse(HTTPResponse):
    """
    Returns response object with output file.
    """
    def __init__(self, location, mime_type=None, filename=None,
                 _range=None, status=200, headers=None):
        headers = headers or {}
        if filename:
            headers["Content-Disposition"] = f'attachment; filename="{filename}"'
        self._range = _range
        self.location = location
        self.filename = filename or os.path.split(self.location)[-1]
        self.mime_type = mime_type or mimetypes.guess_type(self.filename)[0] or "text/plain"
        super(FileResponse, self).__init__('', status, headers, self.mime_type)

    async def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        async with open_async(self.location, mode="rb") as _file:
            if self._range:
                await _file.seek(self._range.start)
                out_stream = await _file.read(self._range.size)
                self.headers["Content-Range"] = "bytes %s-%s/%s" % (
                    self._range.start,
                    self._range.end,
                    self._range.total,
                )
                self.status = 206
            else:
                out_stream = await _file.read()
        self.body = self._encode_body(out_stream)
        return await super(FileResponse, self).output(version, keep_alive, keep_alive_timeout)


class StreamResponse(StreamHTTPResponse):
    """
    Returns response object with output stream.
    """
    def __init__(self, location, mime_type=None, filename=None,
                 _range=None, status=200, headers=None, chunk_size=4096):
        headers = headers or {}
        if filename:
            headers["Content-Disposition"] = f'attachment; filename="{filename}"'
        self._range = _range
        self.location = location
        self.chunk_size = chunk_size
        self.filename = filename or os.path.split(self.location)[-1]
        self.mime_type = mime_type or mimetypes.guess_type(self.filename)[0] or "text/plain"
        super(StreamResponse, self).__init__(None, status, headers, self.mime_type)

    async def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        _file = await open_async(self.location, mode="rb")

        async def _stream_fn(response):
            nonlocal _file
            try:
                if self._range:
                    chunk_size = min((self._range.size, self.chunk_size))
                    await _file.seek(self._range.start)
                    to_send = self._range.size
                    while to_send > 0:
                        content = await _file.read(chunk_size)
                        if len(content) < 1:
                            break
                        to_send -= len(content)
                        await response.write(content)
                else:
                    while True:
                        content = await _file.read(self.chunk_size)
                        if len(content) < 1:
                            break
                        await response.write(content)
            finally:
                await _file.close()
        self.stream_fn = _stream_fn
        return await super(StreamResponse, self).output(version, keep_alive, keep_alive_timeout)



import math
from alita.base import BaseConverter


class StringConverter(BaseConverter):
    regex = "[^/]+"

    def convert(self, value):
        return value

    def to_string(self, value):
        value = str(value)
        assert "/" not in value, "May not contain path seperators"
        assert value, "Must not be empty"
        return value


class PathConverter(BaseConverter):
    regex = ".*"

    def convert(self, value):
        return str(value)

    def to_string(self, value):
        return str(value)


class IntegerConverter(BaseConverter):
    regex = "[0-9]+"

    def convert(self, value):
        return int(value)

    def to_string(self, value):
        value = int(value)
        assert value >= 0, "Negative integers are not supported"
        return str(value)


class FloatConverter(BaseConverter):
    regex = "[0-9]+(.[0-9]+)?"

    def convert(self, value):
        return float(value)

    def to_string(self, value):
        value = float(value)
        assert value >= 0.0, "Negative floats are not supported"
        assert not math.isnan(value), "NaN values are not supported"
        assert not math.isinf(value), "Infinite values are not supported"
        return ("%0.20f" % value).rstrip("0").rstrip(".")


CONVERTER_TYPES = {
    "str": StringConverter(),
    "path": PathConverter(),
    "int": IntegerConverter(),
    "float": FloatConverter(),
}



import os
import sys
import json
import http.cookies
from urllib import parse
from multidict import CIMultiDict
from alita.serve import STATUS_TEXT
from alita.utils import get_request_url
from alita.utils import cached_property, to_unicode, \
    has_message_body, remove_entity_headers, parse_options_header


class JSONSerializer:
    """
    Simple wrapper around json to be used in signing.dumps and
    signing.loads.
    """
    def dumps(self, obj):
        return json.dumps(obj, separators=(',', ':'))

    def loads(self, data):
        return json.loads(data)


class BaseFactory:

    def create_request_object(self, environ):
        raise NotImplementedError

    def create_base_response_class(self):
        raise NotImplementedError

    def create_base_exception_class(self):
        raise NotImplementedError

    def create_exception_handler_object(self):
        raise NotImplementedError

    def create_router_object(self):
        raise NotImplementedError


class BaseRequest(object):
    charset = 'utf-8'
    encoding_errors = 'replace'

    def __init__(self, app, environ, headers=None):
        # Request state
        self.app = app
        self.environ = environ
        self.headers = CIMultiDict(headers or {})

        # Response state
        self._cookies = None
        self._cached_data = None
        self._parsed_content_type = None
        self.disconnected = False
        self.response_complete = False
        self.match_headers()

    def match_headers(self):
        for value in self.environ["headers"]:
            self.headers[value[0].decode()] = value[1].decode()

    @property
    def content_type(self):
        """Like :attr:`content_type`, but without parameters (eg, without
        charset, type etc.) and always lowercase.  For example if the content
        type is ``text/HTML; charset=utf-8`` the mimetype would be
        ``'text/html'``.
        """
        if self.headers.get('content-type'):
            self._parsed_content_type = parse_options_header(self.headers['content-type'])
        else:
            self._parsed_content_type = parse_options_header(self.headers.get('accept', ''))
        return self._parsed_content_type[0].lower()

    @cached_property
    def cookies(self):
        if self._cookies is None:
            cookies = {}
            cookie_header = self.headers.get("cookie")
            if cookie_header:
                cookie = http.cookies.SimpleCookie()
                cookie.load(cookie_header)
                for key, morsel in cookie.items():
                    cookies[key] = morsel.value
            self._cookies = cookies
        return self._cookies

    @cached_property
    def version(self):
        """
        The http request version.
        """
        return self.environ.get("http_version", "1.1")

    @cached_property
    def path(self):
        """
        Requested path as unicode.  This works a bit like the regular path
        info in the WSGI environment but will always include a leading slash,
        even if the URL root is accessed.
        """
        raw_path = self.environ.get('path') or ''
        return '/' + raw_path.lstrip('/')

    @property
    def url_charset(self):
        """
        The charset that is assumed for URLs.  Defaults to the value
        of :attr:`charset`.

        .. versionadded:: 0.6
        """
        return self.charset

    @cached_property
    def full_path(self):
        """
        Requested path as unicode, including the query string.
        """
        return self.path + u'?' + to_unicode(self.query_string, self.url_charset)

    @cached_property
    def body(self):
        """
        The reconstructed current URL as IRI.
        """
        return self.environ.get("body")

    @cached_property
    def url(self):
        """
        The reconstructed current URL as IRI.
        """
        return self.environ.get("url")

    @cached_property
    def base_url(self):
        """
        Like :attr:`url` but without the querystring
        """
        return get_request_url(self, strip_querystring=True)

    @cached_property
    def root_url(self):
        """
        The full URL root (with hostname), this is the application
        root as IRI.
        """
        return get_request_url(self, root_only=True)

    @cached_property
    def host(self):
        """
        Just the host including the port if available.
        """
        return self.environ.get("host")

    @cached_property
    def port(self):
        """
        Just the host including the port if available.
        """
        return self.environ.get("port")

    @cached_property
    def query_string(self):
        """
        Just the request query string.
        """
        return self.environ.get("query_string")

    @cached_property
    def path_params(self):
        """
        Just the request query string.
        """
        return parse.parse_qs(parse.urlsplit(self.full_path).query)

    @cached_property
    def method(self):
        """
        Just the request method string.
        """
        return self.environ.get("method")

    @cached_property
    def scheme(self):
        """
        Just the request scheme string.
        """
        return self.environ.get("scheme")

    @cached_property
    def server(self):
        """
        Just the request server string.
        """
        return self.environ.get("server")

    @cached_property
    def root_path(self):
        """
        Just the request root_path string.
        """
        return self.environ.get("root_path")

    def get_data(self, cache=True, as_text=False):
        rv = getattr(self, '_cached_data', None)
        if rv is None:
            rv = self.body
            if cache:
                self._cached_data = rv
        if as_text:
            rv = rv.decode(self.charset, self.encoding_errors)
        return rv


class BaseResponse:
    __slots__ = ("body", "status", "content_type", "headers", "_cookies", "_protocol")

    charset = 'utf-8'
    max_cookie_size = 4093

    def __init__(self, body=None, status=200, headers=None, content_type="text/plain"):
        self.content_type = content_type
        self.body = self._encode_body(body)
        self.status = status
        self.headers = CIMultiDict(headers or {})
        self._cookies = None
        self._protocol = None

    def set_protocol(self, protocol):
        self._protocol = protocol

    def has_protocol(self):
        return self._protocol is not None

    @staticmethod
    def _encode_body(data):
        try:
            if not isinstance(data, bytes):
                return data.encode()
            return data
        except AttributeError:
            return str(data or "").encode()

    def _parse_headers(self):
        headers = b""
        for name, value in self.headers.items():
            try:
                headers += b"%b: %b\r\n" % (
                    name.encode(),
                    value.encode(self.charset),
                )
            except AttributeError:
                headers += b"%b: %b\r\n" % (
                    str(name).encode(),
                    str(value).encode(self.charset),
                )
        return headers

    def set_cookie(self, key, value="", max_age=None, expires=None, path="/",
                   domain=None, secure=False, httponly=False, samesite=None):
        cookie = http.cookies.SimpleCookie()
        cookie[key] = value
        if max_age is not None:
            cookie[key]["max-age"] = max_age  # type: ignore
        if expires is not None:
            cookie[key]["expires"] = expires  # type: ignore
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite:
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.headers.add('Set-Cookie', cookie_val)

    def delete_cookie(self, key, path="/", domain=None):
        self.set_cookie(key, expires=0, max_age=0, path=path, domain=domain)

    def get_headers(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        timeout_header = b""
        if keep_alive and keep_alive_timeout is not None:
            timeout_header = b"Keep-Alive: %d\r\n" % keep_alive_timeout
        self.headers["Content-Type"] = self.headers.get(
            "Content-Type", self.content_type
        )
        if self.status in (304, 412):
            self.headers = remove_entity_headers(self.headers)
        headers = self._parse_headers()
        if self.status is 200:
            description = b"OK"
        else:
            description = STATUS_TEXT.get(self.status, b"UNKNOWN RESPONSE")

        return (
            b"HTTP/%b %d %b\r\n" b"Connection: %b\r\n" b"%b" b"%b\r\n"
        ) % (
            version.encode(),
            self.status,
            description,
            b"keep-alive" if keep_alive else b"close",
            timeout_header,
            headers
        )

    async def output(self, version="1.1", keep_alive=False, keep_alive_timeout=None):
        if has_message_body(self.status):
            body = self.body
            self.headers["Content-Length"] = self.headers.get(
                "Content-Length", len(self.body)
            )
        else:
            body = b""
        return self.get_headers(
            version,
            keep_alive,
            keep_alive_timeout
        ) + b"%b" % body


class BaseHTTPException(Exception):
    """
    Baseclass for all HTTP exceptions.  This exception can be called as WSGI
    application to render a default error page or you can catch the subclasses
    of it independently and render nicer error messages.
    """

    code = None
    description = None

    def __init__(self, description=None, response=None, code=None):
        super(BaseHTTPException, self).__init__(self)
        if code is not None:
            self.code = code
        if description is not None:
            self.description = description
        self.response = response

    @classmethod
    def wrap(cls, exception, name=None):
        """
        This method returns a new subclass of the exception provided that
        also is a subclass of `BadRequest`.
        """
        class newcls(cls, exception):

            def __init__(self, arg=None, *args, **kwargs):
                cls.__init__(self, *args, **kwargs)
                exception.__init__(self, arg)
        newcls.__module__ = sys._getframe(1).f_globals.get('__name__')
        newcls.__name__ = name or cls.__name__ + exception.__name__
        return newcls

    @property
    def name(self):
        """
        The status name.
        """
        return STATUS_TEXT.get(self.code, 'Unknown Error')

    def __call__(self, environ, start_response):
        """
        Call the exception as WSGI application.

        :param environ: the WSGI environment.
        :param start_response: the response callable provided by the WSGI
                               server.
        """
        response = self.get_response(environ)
        return response(environ, start_response)

    def __str__(self):
        code = self.code if self.code is not None else '???'
        return '%s %s: %s' % (code, self.name, self.description)

    def __repr__(self):
        code = self.code if self.code is not None else '???'
        return "<%s '%s: %s'>" % (self.__class__.__name__, code, self.name)

    def get_description(self, environ=None):
        raise NotImplementedError

    def get_body(self, environ=None):
        raise NotImplementedError

    def get_response(self, environ=None):
        raise NotImplementedError

    def get_headers(self, environ=None):
        raise NotImplementedError


class BaseMiddleware:
    def __init__(self, app):
        self.app = app

    async def process_request(self, request):
        pass

    async def process_response(self, request, response):
        return response


class BaseExceptionHandler:
    def __init__(self, app=None):
        self.app = app

    def process_exception(self, request, exc):
        raise NotImplementedError


class BaseBlueprint:
    def register(self, app, **options):
        raise NotImplementedError

    def add_url_rule(self, rule, view_func, **options):
        raise NotImplementedError

    def before_request(self, f):
        pass

    def after_request(self, f):
        pass


class BaseRoute:
    def match(self, request):
        raise NotImplementedError()

    def url_path_for(self, endpoint, **path_params):
        raise NotImplementedError()


class BaseRouter:
    def match(self, request):
        raise NotImplementedError()

    def add_route(self, path, endpoint, view_func, methods=None):
        raise NotImplementedError()

    def url_path_for(self, endpoint, **path_params):
        raise NotImplementedError()


class BaseConverter:
    regex = ""

    def convert(self, value):
        raise NotImplementedError()

    def to_string(self, value):
        raise NotImplementedError()


class BaseStaticHandler:
    def __init__(self, app):
        self.app = app
        self.static_folder = self.app.static_folder
        self.static_url_path = self.app.static_url_path
        self.send_file_max_age = self.app.send_file_max_age
        if self.static_folder:
            self.app.add_url_rule(
                self.get_static_url_path() + '/<file_name:path>',
                endpoint='static',
                view_func=self.send_static_file
            )

    @property
    def has_static_folder(self):
        return self.static_folder is not None

    def get_static_url_path(self):
        if self.static_url_path is not None:
            return self.static_url_path

        if self.static_folder is not None:
            return '/' + os.path.basename(self.static_folder)

    async def send_static_file(self, request, file_name):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from itertools import repeat

iteritems = lambda d, *args, **kwargs: iter(d.items(*args, **kwargs))


def is_immutable(self):
    raise TypeError('%r objects are immutable' % self.__class__.__name__)


class _Missing(object):

    def __repr__(self):
        return 'no value'

    def __reduce__(self):
        return '_missing'


_missing = _Missing()


class ImmutableListMixin(object):
    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        is_immutable(self)

    def __iadd__(self, other):
        is_immutable(self)
    __imul__ = __iadd__

    def __setitem__(self, key, value):
        is_immutable(self)

    def append(self, item):
        is_immutable(self)
    remove = append

    def extend(self, iterable):
        is_immutable(self)

    def insert(self, pos, value):
        is_immutable(self)

    def pop(self, index=-1):
        is_immutable(self)

    def reverse(self):
        is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        is_immutable(self)


class ImmutableList(ImmutableListMixin, list):
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            list.__repr__(self),
        )


class ImmutableDictMixin(object):
    _hash_cache = None

    @classmethod
    def fromkeys(cls, keys, value=None):
        instance = super(cls, cls).__new__(cls)
        instance.__init__(zip(keys, repeat(value)))
        return instance

    def __reduce_ex__(self, protocol):
        return type(self), (dict(self),)

    def _iter_hashitems(self):
        return iteritems(self)

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(frozenset(self._iter_hashitems()))
        return rv

    def setdefault(self, key, default=None):
        is_immutable(self)

    def update(self, *args, **kwargs):
        is_immutable(self)

    def pop(self, key, default=None):
        is_immutable(self)

    def popitem(self):
        is_immutable(self)

    def __setitem__(self, key, value):
        is_immutable(self)

    def __delitem__(self, key):
        is_immutable(self)

    def clear(self):
        is_immutable(self)


class ImmutableMultiDictMixin(ImmutableDictMixin):
    def __reduce_ex__(self, protocol):
        return type(self), (list(iteritems(self, multi=True)),)

    def _iter_hashitems(self):
        return iteritems(self, multi=True)

    def add(self, key, value):
        is_immutable(self)

    def popitemlist(self):
        is_immutable(self)

    def poplist(self, key):
        is_immutable(self)

    def setlist(self, key, new_list):
        is_immutable(self)

    def setlistdefault(self, key, default_list=None):
        is_immutable(self)


class ImmutableDict(ImmutableDictMixin, dict):
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            dict.__repr__(self),
        )

    def copy(self):
        return dict(self)

    def __copy__(self):
        return self


class UpdateDictMixin(object):

    """Makes dicts call `self.on_update` on modifications.

    .. versionadded:: 0.5

    :private:
    """

    on_update = None

    def calls_update(name):
        def oncall(self, *args, **kw):
            rv = getattr(super(UpdateDictMixin, self), name)(*args, **kw)
            if self.on_update is not None:
                self.on_update(self)
            return rv
        oncall.__name__ = name
        return oncall

    def setdefault(self, key, default=None):
        modified = key not in self
        rv = super(UpdateDictMixin, self).setdefault(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    def pop(self, key, default=_missing):
        modified = key in self
        if default is _missing:
            rv = super(UpdateDictMixin, self).pop(key)
        else:
            rv = super(UpdateDictMixin, self).pop(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    __setitem__ = calls_update('__setitem__')
    __delitem__ = calls_update('__delitem__')
    clear = calls_update('clear')
    popitem = calls_update('popitem')
    update = calls_update('update')
    del calls_update


class CallbackDict(UpdateDictMixin, dict):

    """A dict that calls a function passed every time something is changed.
    The function is passed the dict instance.
    """

    def __init__(self, initial=None, on_update=None):
        dict.__init__(self, initial or ())
        self.on_update = on_update

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            dict.__repr__(self)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import re
from setuptools import setup
from collections import OrderedDict

with io.open('README.md', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('alita/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='alita',
    version=version,
    url='https://github.com/dwpy/alita',
    project_urls=OrderedDict((
        ('Documentation', 'https://dwpy.github.io/alita'),
        ('Code', 'https://github.com/dwpy/alita'),
    )),
    license='BSD',
    author='Dongwei',
    description='A simple Python async framework for building web applications. ',
    long_description=readme,
    packages=[
        'alita',
        'alita.serve',
        'alita.contrib'
    ],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    python_requires='>=3.5',
    install_requires=[
        'bson',
        'attrs',
        'blinker',
        'click',
        'uvloop',
        'multidict',
        'itsdangerous',
        'httptools',
        'websockets',
        'gunicorn',
        'aiofiles',
    ],
    extras_require={
        'dotenv': ['python-dotenv'],
        'docs': [
            'sphinx',
            'pallets-sphinx-themes',
            'sphinxcontrib-log-cabinet',
        ]
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'alita = alita.contrib.command.cli:main',
        ],
    },
)



import json
from alita.base import BaseRequest
from alita.exceptions import BadRequest


class JSONMixin(object):
    """
    Common mixin for both request and response objects to provide JSON
    """
    _cached_json = (Ellipsis, Ellipsis)

    @property
    def is_json(self):
        mt = self.content_type
        return (
            mt == 'application/json'
            or (mt.startswith('application/')) and mt.endswith('+json')
        )

    @property
    def json(self):
        return self.get_json()

    def _get_data_for_json(self, cache):
        return self.get_data(cache=cache)

    def get_json(self, force=False, silent=False, cache=True):
        if cache and self._cached_json[silent] is not Ellipsis:
            return self._cached_json[silent]

        if not (force or self.is_json):
            return None

        data = self._get_data_for_json(cache=cache)

        try:
            rv = json.loads(data)
        except ValueError as e:
            if silent:
                rv = None
                if cache:
                    normal_rv, _ = self._cached_json
                    self._cached_json = (normal_rv, rv)
            else:
                rv = self.on_json_loading_failed(e)
                if cache:
                    _, silent_rv = self._cached_json
                    self._cached_json = (rv, silent_rv)
        else:
            if cache:
                self._cached_json = (rv, rv)
        return rv

    def on_json_loading_failed(self, ex):
        raise BadRequest('Failed to decode JSON object: {0}'.format(ex))


class Request(BaseRequest, JSONMixin):
    route_match = None
    routing_exception = None

    def __init__(self, app, environ, headers=None):
        super(Request, self).__init__(app, environ, headers)
        self.match_request()

    def match_request(self):
        try:
            self.route_match = self.app.router.match(self)
        except self.app.exception_class as ex:
            self.routing_exception = ex
        except Exception as ex:
            self.routing_exception = BadRequest(str(ex))

    @property
    def endpoint(self):
        return self.route_match.endpoint if self.route_match else None

    @property
    def blueprint(self):
        """
        The name of the current blueprint
        """
        if self.route_match and '.' in self.route_match.endpoint:
            return self.route_match.endpoint.rsplit('.', 1)[0]



# -*- coding: utf-8 -*-
from alita.base import BaseBlueprint


class Blueprint(BaseBlueprint):
    _has_registered = False

    def __init__(self, name, url_prefix=None):
        self.name = name
        self.url_prefix = url_prefix
        self.deferred_functions = []
        self.app = None

    @property
    def registered(self):
        return self._has_registered

    def record(self, func):
        if self._has_registered:
            from warnings import warn
            warn('The blueprint was already registered once '
                 'but is getting modified now.  These changes '
                 'will not show up.')
        self.deferred_functions.append(func)
    
    def register(self, app, **options):
        self.app = app
        for option in options:
            if hasattr(self, option):
                setattr(self, option, options[option])
        self._has_registered = True
        for deferred in self.deferred_functions:
            deferred(self)

    def route(self, rule, **options):
        def decorator(f):
            self.add_url_rule(rule, f, **options)
            return f
        return decorator

    def add_app_url_rule(self, rule, view_func, **options):
        endpoint = options.get('endpoint')
        if self.url_prefix is not None:
            if rule:
                rule = '/'.join((
                    self.url_prefix.rstrip('/'), rule.lstrip('/')))
            else:
                rule = self.url_prefix
        if endpoint is None:
            endpoint = self.app.get_endpoint_from_view_func(view_func)
        options['endpoint'] = '%s.%s' % (self.name, endpoint)
        self.app.add_url_rule(rule, view_func, **options)

    def add_url_rule(self, rule, view_func, **options):
        endpoint = options.get('endpoint')
        if endpoint:
            assert '.' not in endpoint, "Blueprint endpoints should not contain dots"
        if view_func and hasattr(view_func, '__name__'):
            assert '.' not in view_func.__name__, "Blueprint view function name should not contain dots"
        self.record(lambda s: s.add_app_url_rule(rule, view_func, **options))

    def endpoint(self, endpoint):
        def decorator(f):
            def register_endpoint():
                self.app.view_functions[endpoint] = f
            self.record(register_endpoint)
            return f
        return decorator

    def before_request(self, f):
        self.record(lambda s: s.app.before_request_funcs.setdefault(self.name, []).append(f))
        return f

    def after_request(self, f):
        self.record(lambda s: s.app.after_request_funcs.setdefault(self.name, []).append(f))
        return f
